const { default: axios } = require("axios");
const { ethers } = require("ethers");

import { TransferType } from "../constants/index";
import * as contracts from "../abis/index";
import { config } from "../config/nets";

class RouterSDK {
    constructor(signerOrProvider, optionalParams) {
        return (async () => {

            // All async code here
            this.signerOrProvider = signerOrProvider;
        
            this.chainId = this.signerOrProvider.provider.network.chainId;
            this.config = config[this.chainId];
    
            this.setParams(optionalParams || {});
            this.routerContract = new ethers.Contract(this.routerAddress, this.routerABI, this.signerOrProvider);
            this.proxyRouterContract = new ethers.Contract(this.proxyRouterAddress, this.proxyRouterABI, this.signerOrProvider);
            let validatorArray = [];
            let initialValidatorVersion;

            try {
                initialValidatorVersion = (await axios.get(`${this.validatorURL}/status`)).data.version;
            } catch (e) {
                initialValidatorVersion = await this._getDefaultConfiguredValidatorVersion();
                if(!initialValidatorVersion) {
                    throw new Error('Validators are not accessible')
                }
            }
           
            const initialValidatorMajorVersion = parseInt(initialValidatorVersion.split('.')[0]);
            const initialValidatorMinorVersion = parseInt(initialValidatorVersion.split('.')[1]);
            const optionalParamsValidatorMajorVersion = parseInt(optionalParams["validatorTargetVersion"].split('.')[0]);
            const optionalParamsValidatorMinorVersion = parseInt(optionalParams["validatorTargetVersion"].split('.')[1]);
            this.axios = new axios.create({ baseURL: this.validatorURL });
            this.validatorVersion = initialValidatorVersion
            // Check for passed validatorTargetVersion
            if((optionalParams &&  optionalParams["validatorTargetVersion"]) && (initialValidatorMajorVersion !== optionalParamsValidatorMajorVersion || !(initialValidatorMinorVersion >= optionalParamsValidatorMinorVersion))) {
                
                if (optionalParams && optionalParams.validators) { 
                    validatorArray = this._getAllValidators(optionalParams.validators)
                } else {
                    validatorArray = this._getAllValidators(this.config.validators)
                }
                await this._setValidatorBasedOnVersion(validatorArray, optionalParams["validatorTargetVersion"])
            }          

            return this; // when done
        })();
    }

    /**
     * @param optionalParams
     * @property routerABI: RouterContractABI 
     * @property proxyRouterABI: proxyRouterContractABI 
     * @property erc20ABI: ERC20Permit 
     * @property routerAddress: address 
     * @property proxyRouterAddress: address 
     * @property validators: object
     * @property threshold: number 
     * @property validatorTargetVersion: string
     */
    setParams(optionalParams) {
        
        if (optionalParams["routerABI"]) {
            this.routerABI = optionalParams.routerABI;
        } else {
            this.routerABI = contracts.routerContract.abi;
        }

        if (optionalParams["proxyRouterABI"]) {
            this.proxyRouterABI = optionalParams.proxyRouterABI;
        } else {
            this.proxyRouterABI = contracts.proxyRouterContract.abi;
        }
        
        if (optionalParams["erc20ABI"]) {
            this.erc20ABI = optionalParams.erc20ABI;
        } else {
            this.erc20ABI = contracts.ERC20Permit.abi;
        }
        
        if (optionalParams["routerAddress"]) {
            this.routerAddress = optionalParams.routerAddress;
        } else {
            this.routerAddress = this.config.contracts.router;
        }

        if (optionalParams["proxyRouterAddress"]) {
            this.proxyRouterAddress = optionalParams.proxyRouterAddress;
        } else {
            this.proxyRouterAddress = this.config.contracts['proxy-router'];
        }
        
        if (optionalParams["validators"]) {
            this.validatorURL = this.getRandomProperty(optionalParams.validators);
        } else {
            this.validatorURL = this.getRandomProperty(this.config.validators);
        }

        let threshold;
        if (optionalParams["threshold"]) {
            threshold = ethers.utils.formatUnits(optionalParams.threshold)
        } else {
            threshold = ethers.utils.formatUnits(this.config.threshold)
        }

        this.tokenAllowanceThreshold = ethers.utils.parseUnits(threshold);
    }

    setValidator(url) {
        this.validatorURL = url;
        this.axios = new axios.create({ baseURL: this.validatorURL });
    }
    
    getRandomProperty(obj) {
        const keys = Object.keys(obj);
        const randomIndex = keys.length * Math.random() << 0;

        return obj[keys[randomIndex]];
    }

    async send(targetChainId, tokenAddress, amount, receiver, feeTokenAddress, bridgedTokenIsNativeToken) {
        const nativeToken = await this.routerContract.wrappedToNativeToken(tokenAddress);
        if(feeTokenAddress) {
            // Check of fee token is native token
            let fee = feeTokenAddress === this.proxyRouterContract.address ? await this.getProxyFee(feeTokenAddress) : 0;
            
            // Check if bridged token is native token
            if(bridgedTokenIsNativeToken) {
                fee = amount.add(fee);
            }
            if(nativeToken.chainId == 0) {
                return await this.proxyRouterContract.populateTransaction.lock(feeTokenAddress, targetChainId, tokenAddress, amount, receiver, {value: fee});
            } else if (nativeToken.chainId !== targetChainId) {
                return await this.proxyRouterContract.populateTransaction.burnAndTransfer(feeTokenAddress, targetChainId, tokenAddress, amount, receiver, {value: fee})
            }

            return await this.proxyRouterContract.populateTransaction.burn(feeTokenAddress, tokenAddress, amount, receiver, {value: fee })
            
        } else {
            if(nativeToken.chainId == 0) {
                return await this.routerContract.populateTransaction.lock(targetChainId, tokenAddress, amount, receiver);
            } else if (nativeToken.chainId !== targetChainId) {
                return await this.routerContract.populateTransaction.burnAndTransfer(targetChainId, tokenAddress, amount, receiver);
            }
    
            return await this.routerContract.populateTransaction.burn(tokenAddress, amount, receiver);
        }
    }

    async sendWithPermit(targetChainId, tokenAddress, amount, receiver, v, r, s, deadline, feeTokenAddress, bridgedTokenIsNativeToken) {
        const nativeToken = await this.routerContract.wrappedToNativeToken(tokenAddress);
        if(feeTokenAddress) {
            // Check of fee token is native token
            let fee = feeTokenAddress === this.proxyRouterContract.address ? await this.getProxyFee(feeTokenAddress) : 0;
           
            // Check if bridged token is native token
            if(bridgedTokenIsNativeToken) {
                fee = amount.add(fee);
            }
            if(nativeToken.chainId == 0) {
                return await this.proxyRouterContract.populateTransaction.lockWithPermit(feeTokenAddress ,targetChainId, tokenAddress, amount, receiver, deadline, v, r, s, {value: fee});
            } else if (nativeToken.chainId !== targetChainId) {
                return await this.proxyRouterContract.populateTransaction.burnAndTransferWithPermit(feeTokenAddress, targetChainId, tokenAddress, amount, receiver, deadline, v, r, s, {value: fee});
            }
    
            return await this.proxyRouterContract.populateTransaction.burnWithPermit(feeTokenAddress, tokenAddress, amount, receiver, deadline, v, r, s, {value: fee});
        } else {
            if(nativeToken.chainId == 0) {
                return await this.routerContract.populateTransaction.lockWithPermit(targetChainId, tokenAddress, amount, receiver, deadline, v, r, s);
            } else if (nativeToken.chainId !== targetChainId) {
                return await this.routerContract.populateTransaction.burnAndTransferWithPermit(targetChainId, tokenAddress, amount, receiver, deadline, v, r, s);
            }
    
            return await this.routerContract.populateTransaction.burnWithPermit(tokenAddress, amount, receiver, deadline, v, r, s);
        }
      
    }

    async claim(transfer) {
        switch(transfer.transferType) {
            case TransferType.LOCK:
                return await this.routerContract.populateTransaction.mint(
                    transfer.nativeChainId === 0 ? transfer.sourceChainId : transfer.nativeChainId,
                    transfer.nativeToken,
                    ethers.utils.toUtf8Bytes(transfer.sourceTx),
                    transfer.amount,
                    transfer.recipient,
                    this.fixSignatures(transfer.signatures),
                    { name: transfer.wrappedTokenName,
                    symbol: transfer.wrappedTokenSymbol,
                    decimals: transfer.wrappedTokenDecimals }
                );
            case TransferType.BURN:
                return await this.routerContract.populateTransaction.unlock(
                    transfer.sourceChainId,
                    ethers.utils.toUtf8Bytes(transfer.sourceTx),
                    transfer.nativeToken,
                    transfer.amount,
                    transfer.recipient,
                    this.fixSignatures(transfer.signatures)
                );
            case TransferType.BURN_AND_TRANSFER:
                return await this.routerContract.populateTransaction.mint(
                    transfer.nativeChainId,
                    transfer.nativeToken,
                    ethers.utils.toUtf8Bytes(transfer.sourceTx),
                    transfer.amount,
                    transfer.recipient,
                    this.fixSignatures(transfer.signatures),
                    { name: transfer.wrappedTokenName,
                    symbol: transfer.wrappedTokenSymbol,
                    decimals: transfer.wrappedTokenDecimals }
                );
        }

        return null;
    }

    async balanceOf(tokenAddress, userAddress) {
        let balance;
        let tokenDecimals;
        if(tokenAddress === this.proxyRouterContract.address) {
            balance = await this.signerOrProvider.getBalance();
            tokenDecimals = await this.proxyRouterContract.decimals();
        } else {
            const tokenContract = new ethers.Contract(tokenAddress, this.erc20ABI, this.signerOrProvider);
            balance = await tokenContract.balanceOf(userAddress);
            tokenDecimals = await tokenContract.decimals();
        }
        
        return { balance, tokenDecimals };
    }

    async allowance(tokenAddress, userAddress) {
        const tokenContract = new ethers.Contract(tokenAddress, this.erc20ABI, this.signerOrProvider);
        const tokenDecimals = await tokenContract.decimals();
        const allowance = await tokenContract.allowance(userAddress, this.routerContract.address);
        return { allowance, tokenDecimals };
    }

    async isTokenApproved(tokenAddress, userAddress, bridgedTokenIsNativeTokenOrNonDefaultFee, threshold /** optional */) {
        const allowanceThreshold = threshold || this.tokenAllowanceThreshold;
        const tokenContract = new ethers.Contract(tokenAddress, this.erc20ABI, this.signerOrProvider);
        const currentAllowance = await tokenContract.allowance(userAddress, bridgedTokenIsNativeTokenOrNonDefaultFee ? this.proxyRouterContract.address : this.routerContract.address);
        return currentAllowance.gte(allowanceThreshold);
    }

    async approve(tokenAddress, amount, bridgedTokenIsNativeTokenOrNonDefaultFee) {
        const tokenContract = new ethers.Contract(tokenAddress, this.erc20ABI, this.signerOrProvider);
        return await tokenContract.populateTransaction.approve(bridgedTokenIsNativeTokenOrNonDefaultFee ? this.proxyRouterContract.address : this.routerContract.address, amount);
    }

    async approveMaxInt(tokenAddress, bridgedTokenIsNativeTokenOrNonDefaultFee) {
        return this.approve(tokenAddress, ethers.constants.MaxUint256, bridgedTokenIsNativeTokenOrNonDefaultFee);
    }

    async getServiceFee() {
        return await this.routerContract.serviceFee();
    }

    async getExternalFee() {
        return await this.routerContract.externalFee();
    }

    async getClaims(userAddress, targetChainId, page = 'undefined') {
        const transfers = await this.axios(`/getTransfers?targetChainId=${targetChainId}&recipient=${userAddress}&page=${page}`, {responseType: "json"});
        return { claims: transfers.data.transfers, totalCount: transfers.headers['x-total-count']}
    }

    async getValidatorVersion() {
        const versionRequest = await this.axios('/status', {responseType: "json"});
        return versionRequest.data.version;
    }

    async getWrappedTokens(chainId) {
        const wrappedTokens = await this.axios(`/getWrappedTokens?chainId=${chainId}`, {responseType: "json"});
        return wrappedTokens.data.wrappedTokens;
    }

    async isClaimed(transfer) {
        let encodedData;
        switch(transfer.transferType) {
            case TransferType.BURN:
                encodedData = ethers.utils.defaultAbiCoder.encode(
                    [
                        "uint8",
                        "uint8",
                        "bytes",
                        "address",
                        "uint256",
                        "bytes"
                    ],
                    [
                        transfer.sourceChainId,
                        transfer.targetChainId,
                        ethers.utils.toUtf8Bytes(transfer.sourceTx),
                        transfer.recipient,
                        transfer.amount,
                        transfer.nativeToken
                    ]
                );
                break;
            case TransferType.LOCK:
                // Backwards compatibility fix
                if (transfer.nativeChainId === 0) {
                    transfer.nativeChainId = transfer.sourceChainId;
                }
                encodedData = ethers.utils.defaultAbiCoder.encode(
                    [
                        "uint8",
                        "uint8",
                        "bytes",
                        "address",
                        "uint256",
                        "bytes",
                        "string",
                        "string",
                        "uint8"
                    ],
                    [
                        transfer.nativeChainId,
                        transfer.targetChainId,
                        ethers.utils.toUtf8Bytes(transfer.sourceTx),
                        transfer.recipient,
                        transfer.amount,
                        transfer.nativeToken,
                        transfer.wrappedTokenName,
                        transfer.wrappedTokenSymbol,
                        transfer.wrappedTokenDecimals
                    ]
                );
                break;
            case TransferType.BURN_AND_TRANSFER:
                encodedData = ethers.utils.defaultAbiCoder.encode(
                    [
                        "uint8",
                        "uint8",
                        "bytes",
                        "address",
                        "uint256",
                        "bytes",
                        "string",
                        "string",
                        "uint8"
                    ],
                    [
                        transfer.nativeChainId,
                        transfer.targetChainId,
                        ethers.utils.toUtf8Bytes(transfer.sourceTx),
                        transfer.recipient,
                        transfer.amount,
                        transfer.nativeToken,
                        transfer.wrappedTokenName,
                        transfer.wrappedTokenSymbol,
                        transfer.wrappedTokenDecimals
                    ]
                );
                break;
        }
        
        const hashMsg = ethers.utils.keccak256(encodedData);
        const hashData = ethers.utils.arrayify(hashMsg);

        return await this.routerContract.hashesUsed(
                transfer.transferType == TransferType.BURN ? transfer.sourceChainId : transfer.nativeChainId, 
                ethers.utils.hashMessage(hashData)
            );
    }

    async getTokenName(tokenAddress) {
        const tokenContract = new ethers.Contract(tokenAddress, this.erc20ABI, this.signerOrProvider);
        return await tokenContract.name();
    }

    async getTokenSymbol(tokenAddress) {
        const tokenContract = new ethers.Contract(tokenAddress, this.erc20ABI, this.signerOrProvider);
        return await tokenContract.symbol();
    }

    async getTokenDecimals(tokenAddress) {
        const tokenContract = new ethers.Contract(tokenAddress, this.erc20ABI, this.signerOrProvider);
        return await tokenContract.decimals();
    }

    async getTokenDetails(tokenAddress) {
        const tokenContract = new ethers.Contract(tokenAddress, this.erc20ABI, this.signerOrProvider);
        return {
            name: await tokenContract.name(),
            symbol: await tokenContract.symbol(),
            decimals: await tokenContract.decimals()
        }
    }

    async getWrappedTokenDetails(nativeTokenAddress, nativeChainId) {
        const wrappedTokenAddress = await this.routerContract.nativeToWrappedToken(nativeChainId, nativeTokenAddress);
        const wrappedTokenDetails = await this.getTokenDetails(wrappedTokenAddress);
        return { address: wrappedTokenAddress, ...wrappedTokenDetails };
    }

    async getNativeTokenByWrappedAddress(wrappedTokenAddress) {
        return await this.routerContract.wrappedToNativeToken(wrappedTokenAddress);
    }

    async getALBTToken() {
        const albtAddress = await this.routerContract.albtToken();
        const albtDetails = await this.getTokenDetails(albtAddress);
        return { address: albtAddress, ...albtDetails };
    }

    async getProxyFee(address) {
        return await this.proxyRouterContract.feeAmountByToken(address)
    }

    fixSignatures(signatures) {
        const fixedSignatures = [];
        for(let i = 0; i < signatures.length; i++) {
            fixedSignatures[i] = signatures[i].startsWith("0x") ? signatures[i] : "0x" + signatures[i];
        }
    
        return fixedSignatures;
    }

    _getAllValidators(validators) {
        let validatorArray = [];
        for (const key in validators) {
            if (Object.hasOwnProperty.call(validators, key)) {
                const element = validators[key];
                validatorArray.push(element)
            }
        }
        return validatorArray;
    }

    async _setValidatorBasedOnVersion(validatorArray, checkedVersion) {
        const checkedMajorVersion = parseInt(checkedVersion.split('.')[0]);
        const checkedMinorVersion = parseInt(checkedVersion.split('.')[1]);
        for (let index = 0; index < validatorArray.length; index++) {
            let version
            let majorVersion
            let minorVersion
            try {
                version = (await axios.get(`${validatorArray[index]}/status`)).data.version;
                majorVersion =  parseInt(version.split('.')[0]);
                minorVersion = parseInt(version.split('.')[1]);
            } catch (e) {
                continue;
            }
            if(majorVersion === checkedMajorVersion && minorVersion >= checkedMinorVersion) {
                this.validatorVersion = version;
                this.setValidator(validatorArray[index])
                break;
            }
        }   
    }

    async _getDefaultConfiguredValidatorVersion() {
        const validatorArray = this._getAllValidators(this.config.validators);
        for (let index = 0; index < validatorArray.length; index++) {
            try {
                let version = (await axios.get(`${validatorArray[index]}/status`)).data.version;
                this.setValidator(validatorArray[index])
                return version;
            } catch (e) {
                continue;
            }
        }   
    }

}

export { RouterSDK };